import copy
from collections import namedtuple
import operator
import math


ProxyOperation = namedtuple('ProxyOperation', ['func', 'args', 'kwargs'], defaults=[(), {}])


def _operator_call(obj, /, *args, **kwargs): return obj(*args, **kwargs)
def _operator_radd(a, b): return b + a
def _operator_rsub(a, b): return b - a
def _operator_rmul(a, b): return b * a
def _operator_rmatmul(a, b): return b @ a
def _operator_rtruediv(a, b): return b / a
def _operator_rfloordiv(a, b): return b // a
def _operator_rmod(a, b): return b % a
def _operator_rdivmod(a, b): return divmod(b, a)
def _operator_rpow(a, b): return b ** a
def _operator_rlshift(a, b): return b << a
def _operator_rrshift(a, b): return b >> a
def _operator_rand(a, b): return b & a
def _operator_rxor(a, b): return b ^ a
def _operator_ror(a, b): return b | a


def _op_appender(func):
    def _wrapped(proxy, *args, **kwargs):
        return Proxy.append_op(proxy, ProxyOperation(func, args, kwargs))
    return _wrapped


def _format_op(text, op):
    fargs = [repr(arg) for arg in op.args]
    fargs += [f'{key}={value!r}' for key, value in op.kwargs.items()]
    fargs = ', '.join(fargs)
    mapping = {
        operator.lt: '{text} < {fargs}',
        operator.le: '{text} <= {fargs}',
        operator.eq: '{text} == {fargs}',
        operator.ne: '{text} != {fargs}',
        operator.gt: '{text} > {fargs}',
        operator.ge: '{text} >= {fargs}',

        getattr: '{text}.{args[0]!s}', #Use original arg as string
        operator.getitem: '{text}[{fargs}]',
        operator.contains: '{fargs} in {text}',
        _operator_call: '{text}({fargs})',

        operator.add: '{text} + {fargs}',
        operator.sub: '{text} - {fargs}',
        operator.mul: '{text} * {fargs}',
        operator.matmul: '{text} @ {fargs}',
        operator.truediv: '{text} / {fargs}',
        operator.floordiv: '{text} // {fargs}',
        operator.mod: '{text} % {fargs}',
        divmod: 'divmod({text}, {fargs})',
        operator.pow: '{text} ** {fargs}',
        operator.lshift: '{text} << {fargs}',
        operator.rshift: '{text} >> {fargs}',
        operator.and_: '{text} & {fargs}',
        operator.xor: '{text} ^ {fargs}',
        operator.or_: '{text} | {fargs}',

        _operator_radd: '{fargs} + {text}',
        _operator_rsub: '{fargs} - {text}',
        _operator_rmul: '{fargs} * {text}',
        _operator_rmatmul: '{fargs} @ {text}',
        _operator_rtruediv: '{fargs} / {text}',
        _operator_rfloordiv: '{fargs} // {text}',
        _operator_rmod: '{fargs} % {text}',
        _operator_rdivmod: 'divmod({fargs}, {text})',
        _operator_rpow: '{fargs} ** {text}',
        _operator_rlshift: '{fargs} << {text}',
        _operator_rrshift: '{fargs} >> {text}',
        _operator_rand: '{fargs} & {text}',
        _operator_rxor: '{fargs} ^ {text}',
        _operator_ror: '{fargs} | {text}',

        operator.neg: '-{text}',
        operator.pos: '+{text}',
        operator.abs: 'abs({text})',
        operator.invert: '~{text}',

        complex: 'complex({text})',
        int: 'int({text})',
        float: 'float({text})',

        operator.index: 'operator.index({text})',

        float: 'round({text})',
        math.trunc: 'math.trunc({text})',
        math.floor: 'math.floor({text})',
        math.ceil: 'math.ceil({text})',
    }
    template = mapping.get(op.func)
    if template is None:
        return f'{text}.{op.func!s}({fargs})'
    else:
        return template.format(text=text, fargs=fargs, args=op.args)


class ProxyMeta(type):
    def __instancecheck__(cls, instance):
        type_ = type(instance)
        if not issubclass(type_, Proxy):
            return False
        elif issubclass(cls, AssignableProxy):
            return cls.is_assignable(instance) and not issubclass(type_, OutputProxy)
        elif issubclass(cls, InputProxy):
            return not issubclass(type_, OutputProxy)
        elif issubclass(cls, LambdaProxy):
            return cls.is_lambda(instance)
        else:
            return issubclass(type_, cls)


    def append_op(cls, proxy, op):
        return proxy.__class__(*proxy.__operations__, op)

    def instance_in(cls, iterable):
        return any(isinstance(obj, cls) for obj in iterable)    

    def single_type_in(cls, iterable):
        return len(set(type(obj) for obj in iterable if isinstance(obj, cls))) <= 1

    def is_assignable(cls, proxy):
        return all(
            op.func in [getattr, operator.getitem]
            for op in proxy.__operations__
        )

    def is_dummy(cls, proxy):
        if len(proxy.__operations__) != 1:
            return False

        op = proxy.__operations__[0]
        return op.func is getattr and op.args[0] == '_'

    def is_skip(cls, proxy):
        if len(proxy.__operations__) != 1:
            return False

        op = proxy.__operations__[0]
        return op.func in [getattr, operator.getitem] and op.args[0] == 'skip'

    def is_lambda(cls, proxy):
        if len(proxy.__operations__) != 1:
            return False

        op = proxy.__operations__[0]
        return op.func is operator.getitem and callable(op.args[0])

    def apply(cls, obj, args):
        if isinstance(args, dict):
            if cls.instance_in(args.values()):
                return {
                    key: cls.get_value(value, obj) if isinstance(value, cls) else value
                    for key, value in args.items()
                }
            else:
                return args
        elif isinstance(args, (list, tuple)):
            if cls.instance_in(args):
                return args.__class__(
                    cls.get_value(arg, obj) if isinstance(arg, cls) else arg
                    for arg in args
                )
            else:
                return args
        else:
            return cls.get_value(args, obj) if isinstance(args, cls) else args

    def get_value(cls, proxy, obj):
        if not isinstance(proxy, Proxy):
            return proxy
        elif isinstance(proxy, LambdaProxy):
            op = proxy.__operations__[0]
            return op.args[0](obj)
        else:
            orig_obj = obj
            for op in proxy.__operations__:
                args = type(proxy).apply(orig_obj, op.args)
                kwargs = type(proxy).apply(orig_obj, op.kwargs)
                if isinstance(op.func, str):
                    func = getattr(obj, op.func)
                    obj = func(*args, **kwargs)
                else:
                    obj = op.func(obj, *args, **kwargs)
            return obj

    def set_value(cls, proxy, obj, value):
        assert isinstance(proxy, AssignableProxy), "Cannot assign to this Proxy instance"
        if not proxy.__operations__:
            return value
        else:
            orig_obj = obj
            for op in proxy.__operations__[:-1]:
                if isinstance(op.func, str):
                    func = getattr(obj, op.func)
                    obj = func(*op.args, **op.kwargs)
                else:
                    obj = op.func(obj, *op.args, **op.kwargs)

            op = proxy.__operations__[-1]
            if op.func is getattr:
                setattr(obj, op.args[0], value)
            elif op.func is operator.getitem:
                obj[op.args[0]] = value

            return orig_obj


class Proxy(metaclass=ProxyMeta):
    def __init__(self, *operations):
        self.__operations__ = operations

    def __hash__(self):
        hashable_funcs = [getattr, operator.getitem]
        if not isinstance(self, AssignableProxy):
            unhashable_ops = ', '.join(
                op.func for op in self.__operations__
                if op.func not in hashable_funcs
            )
            return TypeError(f'unhashable operations in instance: {unhashable_ops}')
        else:
            return hash(tuple(
                (str(op.func), op.args)
                for op in self.__operations__
                if op.func in hashable_funcs
            ))

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__operations__})'

    def __str__(self):
        #result = '<...>'
        result = self.__class__.__name__[0]
        for op in self.__operations__:
            result = _format_op(result, op)
        return result

    __lt__ = _op_appender(operator.lt)
    __le__ = _op_appender(operator.le)
    __eq__ = _op_appender(operator.eq)
    __ne__ = _op_appender(operator.ne)
    __gt__ = _op_appender(operator.gt)
    __ge__ = _op_appender(operator.ge)

    __getattr__ = _op_appender(getattr)
    __call__ = _op_appender(_operator_call)
    __getitem__ = _op_appender(operator.getitem)
    __contains__ = _op_appender(operator.contains)

    __add__ = _op_appender(operator.add)
    __sub__ = _op_appender(operator.sub)
    __mul__ = _op_appender(operator.mul)
    __matmul__ = _op_appender(operator.matmul)
    __truediv__ = _op_appender(operator.truediv)
    __floordiv__ = _op_appender(operator.floordiv)
    __mod__ = _op_appender(operator.mod)
    __divmod__ = _op_appender(divmod)
    __pow__ = _op_appender(operator.pow)
    __lshift__ = _op_appender(operator.lshift)
    __rshift__ = _op_appender(operator.rshift)
    __and__ = _op_appender(operator.and_)
    __xor__ = _op_appender(operator.xor)
    __or__ = _op_appender(operator.or_)

    __radd__ = _op_appender(_operator_radd)
    __rsub__ = _op_appender(_operator_rsub)
    __rmul__ = _op_appender(_operator_rmul)
    __rmatmul__ = _op_appender(_operator_rmatmul)
    __rtruediv__ = _op_appender(_operator_rtruediv)
    __rfloordiv__ = _op_appender(_operator_rfloordiv)
    __rmod__ = _op_appender(_operator_rmod)
    __rdivmod__ = _op_appender(_operator_rdivmod)
    __rpow__ = _op_appender(_operator_rpow)
    __rlshift__ = _op_appender(_operator_rlshift)
    __rrshift__ = _op_appender(_operator_rrshift)
    __rand__ = _op_appender(_operator_rand)
    __rxor__ = _op_appender(_operator_rxor)
    __ror__ = _op_appender(_operator_ror)

    __iadd__ = _op_appender(operator.iadd)
    __isub__ = _op_appender(operator.isub)
    __imul__ = _op_appender(operator.imul)
    __imatmul__ = _op_appender(operator.imatmul)
    __itruediv__ = _op_appender(operator.itruediv)
    __ifloordiv__ = _op_appender(operator.ifloordiv)
    __imod__ = _op_appender(operator.imod)
    __ipow__ = _op_appender(operator.ipow)
    __ilshift__ = _op_appender(operator.ilshift)
    __irshift__ = _op_appender(operator.irshift)
    __iand__ = _op_appender(operator.iand)
    __ixor__ = _op_appender(operator.ixor)
    __ior__ = _op_appender(operator.ior)

    __neg__ = _op_appender(operator.neg)
    __pos__ = _op_appender(operator.pos)
    __abs__ = _op_appender(operator.abs)
    __invert__ = _op_appender(operator.invert)

    __complex__ = _op_appender(complex)
    __int__ = _op_appender(int)
    __float__ = _op_appender(float)

    __index__ = _op_appender(operator.index)

    __round__ = _op_appender(round)
    __trunc__ = _op_appender(math.trunc)
    __floor__ = _op_appender(math.floor)
    __ceil__ = _op_appender(math.ceil)


class AbstractProxy(Proxy):
    def __new__(cls, *args, **kwargs):
        raise TypeError(f"'{cls.__name__}' is an abstract class and may not be instantiated")


class AssignableProxy(AbstractProxy):
    pass


class InputProxy(AbstractProxy):
    pass


class LambdaProxy(AbstractProxy):
    pass


class StreamProxy(Proxy):
    pass


class ItemProxy(Proxy):
    pass


class OutputProxy(Proxy):
    pass



from . import _core
from pathlib import Path as _Path
import os
import shutil


@_core.Op.using(_core.I.path) >> _core.S.path
def list_dir(path, extensions=None, recursive=False):
    if extensions is None:
        extensions = ['.*']

    path = _Path(path)
    if path.is_dir():
        glob = path.rglob if recursive else path.glob
        for ext in extensions:
            yield from glob('*' + ext)
    else:
        yield path


@_core.Op.using(_core.I.path) >> _core.I.path
def change_dir(cur_path, new_dir):
    return _Path(new_dir) / _Path(cur_path).name


@_core.Op.using(_core.I.path) >> _core.I.path
def change_ext(cur_path, new_ext):
    return _Path(cur_path).with_suffix(new_ext)


@_core.Op.using(_core.O.path, _core.O.skip)
def move_skipped(source, reason):
    skip_dir_name = 'skip'
    if reason:
        skip_dir = _Path(source).parent / skip_dir_name
        target = skip_dir / reason / source.name
        os.makedirs(skip_dir, exist_ok=True)
        shutil.move(source, target)



from . import _core
import numpy as np

#@_core.Op.using(_core.S) >> _core.S
#def progress(stream):
#    #TODO: FIX
#    import tqdm
#    for item in tqdm.tqdm(list(stream)):
#        yield item

@_core.Op >> _core.I.skip
def cond_size(arr, min_size=None, max_size=None):
    if min_size and any(dim < lim for dim, lim in zip(arr.shape, min_size) if lim != -1):
        return 'too_small'
    elif max_size and any(dim > lim for dim, lim in zip(arr.shape, max_size) if lim != -1):
        return 'too_large'
    else:
        return None


@_core.Op.using(_core.S) >> _core.S
def head(stream, n=5):
    for index, item in enumerate(stream):
        if index >= n:
            break
        yield item


@_core.Op.using(_core.S) >> _core.S
def fields(stream, keep=None, drop=None):
    assert bool(keep) != bool(drop), "Must provide either 'keep' or 'remove'"
    if isinstance(keep, str):
        keep = [keep]
    if isinstance(drop, str):
        drop = [drop]

    for item in stream:
        for key in list(item):
            if keep and key not in keep and key != 'skip':
                delattr(item, key)
            elif drop and key in drop:
                delattr(item, key)
        yield item


@_core.Op.using(_core.S) >> _core.S
def purge_skipped(stream):
    for item in stream:
        if not item.skip:
            yield item


@_core.Op
def scale_array(arr, source_range=None, target_range=None, clip=False):
    arr_min = arr.min()
    arr_max = arr.max()
    arr_range = arr_max - arr_min

    src_min, src_max = _core.Value.to_abs(source_range, arr_min, arr_max)
    tgt_min, tgt_max = _core.Value.to_abs(target_range, arr_min, arr_max)
    if src_min is None: src_min = arr_min
    if src_max is None: src_max = arr_max
    if tgt_min is None: tgt_min = arr_min
    if tgt_max is None: tgt_max = arr_max

    assert src_max >= src_min, f'{src_max} is less than {src_min}'
    assert tgt_max >= tgt_min, f'{tgt_max} is less than {tgt_min}'
    arr = tgt_min + (tgt_max - tgt_min) * (arr - src_min) / (src_max - src_min)
    if clip:
        arr = arr.clip(tgt_min, tgt_max)
    return arr


@_core.Op
def clip_array(arr, min_value, max_value):
    arr_min = arr.min()
    arr_max = arr.max()

    min_value = _core.Value.to_abs(min_value, arr_min, arr_max)
    max_value = _core.Value.to_abs(max_value, arr_min, arr_max)

    assert max_value >= min_value, f'{max_value} is less than {min_value}'
    return arr.clip(min_value, max_value)



class Stream:
    def __init__(self, iterable):
        self._iterable = iter(iterable)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._iterable)


class Item(dict):
    __default__ = None

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            return self.__default__

    def __setitem__(self, key, value):
        super().__setitem__(key, value)

    def __delitem__(self, key):
        try:
            super().__delitem__(key)
        except KeyError:
            pass

    def __getattr__(self, name):
        try:
            return super().__getitem__(name)
        except KeyError:
            if name.startswith('__') and name.endswith('__'):
                raise AttributeError(name)
            else:
                return self.__default__

    def __setattr__(self, name, value):
        try:
            super().__getattribute__(name)
        except AttributeError:
            self.__setitem__(name, value)
        else:
            super().__setattr__(name, value)

    def __delattr__(self, name):
        try:
            super().__getattribute__(name)
        except AttributeError:
            self.__delitem__(name)
        else:
            super().__delattr__(name)



from ._proxy import *
from ._stream_item import *
import copy


class Operation:
    __slots__ = ('func', 'args', 'kwargs', 'output')

    def __init__(self, func, *args, **kwargs):
        output = kwargs.pop('__output', None)

        assert callable(func), "func must be callable"
        assert Proxy.single_type_in(args + tuple(kwargs.values())), "Proxies must all be of same type"

        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.output = OutputMapping(output)

    def __repr__(self):
        args = [repr(arg) for arg in self.args]
        args += [f'{key}={value!r}' for key, value in self.kwargs.items()]
        return f'{self.func.__name__}({", ".join(args)}) >> {self.output}'

    def _get_ops(self):
        if hasattr(self, 'operations'):
            return list(self.operations)
        else:
            return [self]

    def __call__(self, *args, **kwargs):
        kwargs = {**self.kwargs, **kwargs}
        return self.using(*self.args, *args, **kwargs)

    def using(self, *args, **kwargs):
        kwargs.setdefault('__output', self.output)
        return self.__class__(self.func, *args, **kwargs)

    def process(self, *args, **kwargs):
        stream, *args = args
        if not isinstance(stream, Stream):
            stream = Stream([stream])
        new_kwargs = {**self.kwargs, **kwargs}
        return Stream(self._call_gen(stream, *self.args, *args, **new_kwargs))

    def _call_gen(self, stream, *args, **kwargs):
        item_args = ItemProxy.instance_in((self.func,) + args + tuple(kwargs.values()))
        #TODO: Fix hack
        obj_args = OutputProxy.instance_in((self.func,) + args + tuple(kwargs.values()))
        if item_args or obj_args:
            for item in stream:
                if not obj_args and item.skip:
                    yield item
                else:
                    new_args = Proxy.apply(item, args)
                    new_kwargs = Proxy.apply(item, kwargs)
                    result = Proxy.get_value(self.func, item) if isinstance(self.func, Proxy) else self.func
                    if callable(result):
                        result = result(*new_args, **new_kwargs)
                    yield from self.output.gen_result(item, result)
        else:
            args = Proxy.apply(stream, args)
            kwargs = Proxy.apply(stream, kwargs)
            func = Proxy.get_value(self.func, stream) if isinstance(self.func, Proxy) else self.func
            result = self.func(*args, **kwargs)
            yield from self.output.gen_result(stream, result)

    def __rshift__(self, other):
        kwargs = self.kwargs.copy()
        kwargs['__output'] = other
        return self.__class__(self.func, *self.args, **kwargs)

    def __or__(self, other):
        ops = self._get_ops()
        if isinstance(other, Operation):
            ops += other._get_ops()
        else:
            return NotImplemented

        return SequentialOperation(ops)


class SequentialOperation(Operation):
    def __init__(self, operations):
        self.operations = operations

    def __repr__(self):
        return f'{self.__class__.__name__}({self.operations!r})'

    def __call__(self, *args, **kwargs):
        first_op, *other_ops, last_op = self.operations
        first_op = first_op(*args, **kwargs)
        return self.__class__([first_op, *other_ops])

    def process(self, *args, **keywords):
        ops = iter(self.operations)
        op = next(ops)
        #Execute first operation with passed args and keywords.
        stream = op.process(*args, **keywords)
        #Then, execute all other operations with the preceding one as input.
        for op in ops:
            stream = op.process(stream)

        return stream


class OutputMapping(dict):
    def __init__(self, mapping):
        if mapping is None:
            mapping = {}
        elif isinstance(mapping, dict):
            pass
        elif isinstance(mapping, (list, tuple)):
            mapping = {
                proxy: OutputProxy()[index]
                for index, proxy in enumerate(mapping)
            }
        else:
            mapping = {mapping: OutputProxy()}

        assert all(isinstance(key, AssignableProxy) for key in mapping.keys()), "All mapping keys must be of type AssignableProxy"
        assert all(isinstance(value, OutputProxy) for value in mapping.values()), "All mapping values must be of type OutputProxy"
        super().__init__(mapping)

    def gen_result(self, obj, result):
        if ItemProxy.instance_in(self):
            new_item = obj if isinstance(obj, Item) else Item()
            yield self.apply(new_item, result)
        elif StreamProxy.instance_in(self):
            for value in result:
                new_item = copy.deepcopy(obj) if isinstance(obj, Item) else Item()
                yield self.apply(new_item, value)
        elif isinstance(obj, Stream):
            yield from obj
        else:
            yield obj

    def apply(self, obj, result):
        for key, value in self.items():
            if Proxy.is_dummy(key):
                continue
            assign_val = Proxy.get_value(value, result)
            if Proxy.is_skip(key) and not assign_val:
                continue
            obj = Proxy.set_value(key, obj, assign_val)
        return obj


class OperationDecorator:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, func, *args, **kwargs):
        kwargs = {**self.kwargs, **kwargs}
        return Operation(func, *self.args, *args, **kwargs)

    def using(self, *args, **kwargs):
        kwargs = {**self.kwargs, **kwargs}
        return self.__class__(*self.args, *args, **kwargs)

    def __getitem__(self, func):
        return self(func)

    def __rshift__(self, other):
        kwargs = self.kwargs.copy()
        kwargs['__output'] = other
        return self.__class__(*self.args, **kwargs)



from ._stream_op import *
from ._stream_item import *
from ._proxy import *
from ._dims import *


S = StreamProxy()
I = ItemProxy()
O = OutputProxy()


Op = OperationDecorator()
Cond = Operation(lambda cond, reason: reason if cond else None) >> I.skip

Rel = RelativeValue(1.)



from . import _core
from transformers import CLIPSegProcessor, CLIPSegForImageSegmentation
import cv2
import numpy as np


_clipseg_transform = None
def _get_clipseg_transform(model_name='CIDAS/clipseg-rd64-refined'):
    global _clipseg_transform
    if _clipseg_transform is None:
        _clipseg_transform = CLIPSegProcessor.from_pretrained(model_name)
    return _clipseg_transform


_clipseg_model = None
def _get_clipseg_model(model_name='CIDAS/clipseg-rd64-refined'):
    global _clipseg_model
    if _clipseg_model is None:
        _clipseg_model = CLIPSegForImageSegmentation.from_pretrained(model_name)
    return _clipseg_model


def get_attention(transform, model, image, captions):
    inputs = transform(text=captions, images=[image] * len(captions), padding=True, return_tensors='pt')
    logits = model.forward(**inputs).logits
    if len(captions) <= 1:
        logits = logits.unsqueeze(0)

    return logits.detach().numpy()


@_core.Op.using(_core.I.image) >> _core.I.attention
def generate_attention(image, captions):
    if isinstance(captions, str):
        captions = {captions: 1.0}
    elif not isinstance(captions, dict):
        captions = {c: 1.0 for c in captions}

    caption_texts = list(captions.keys())
    masks = get_attention(_get_clipseg_transform(), _get_clipseg_model(), image, caption_texts)

    if len(captions) > 1:
        pos = np.zeros(masks.shape[1:])
        neg = np.zeros(masks.shape[1:])
        for layer, weight in zip(masks, captions.values()):
            pos = np.add(pos, weight * np.maximum(layer, 0.0))
            neg = np.add(neg, weight * np.minimum(layer, 0.0))
        mask = np.where(pos > 0.0, pos, neg)
    else:
        mask = masks[0]

    return cv2.resize(mask, (image.shape[1], image.shape[0]))



from . import _core
import cv2
import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
import os
from pathlib import Path


def _scale_rect(rect, target, max_size):
    if len(rect) == 4:
        x0, y0, x1, y1 = rect
    else:
        x0, y0 = rect
        x1, y1 = rect
    w = x1 - x0
    h = y1 - y0
    cx = x0 + w / 2
    cy = y0 + h / 2

    tw, th = target
    mw, mh = max_size

    sx = max(1., w / tw)
    sy = max(1., h / th)
    s = max(sx, sy)

    sx = min(s, mw / tw)
    sy = min(s, mh / th)
    s = min(sx, sy)
    nw = s * tw
    nh = s * th

    nx0 = cx - nw / 2
    ny0 = cy - nh / 2
    nx1 = cx + nw / 2
    ny1 = cy + nh / 2

    if nx0 < 0:
        nx1 += 0 - nx0
        nx0 = 0
    if nx1 > mw:
        nx0 -= nx1 - mw
        nx1 = mw
    if ny0 < 0:
        ny1 += 0 - ny0
        ny0 = 0
    if ny1 > mh:
        ny0 -= ny1 - mh
        ny1 = mh

    return (int(nx0), int(ny0), int(nx1), int(ny1))


@_core.Op.using(_core.I.path) >> _core.I.image
def open_image(path):
    return cv2.imread(str(path))


@_core.Op.using(_core.I.image, _core.I.path)
def save_image(image, path):
    os.makedirs(path.parent, exist_ok=True)
    cv2.imwrite(str(path), image)


@_core.Op.using(_core.O.image, _core.O.path, _core.O.skip)
def save_skipped_image(image, path, reason):
    if reason:
        path = path.parent / 'skip' / reason / path.name
        os.makedirs(path.parent, exist_ok=True)
        cv2.imwrite(str(path), image)


@_core.Op.using(_core.I.image) >> _core.I.focus
def calc_focus(image):
    grayscale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    return cv2.Laplacian(grayscale, cv2.CV_64F).var()


@_core.Op.using(_core.I.image) >> _core.I.lightness
def calc_lightness(image):
    hls = cv2.cvtColor(image, cv2.COLOR_BGR2HLS)
    return cv2.mean(hls)[1] #Get middle channel


@_core.Op.using(_core.I.image) >> _core.I.collage
def calc_collage(im, ignore_padding=0):
    if ignore_padding:
        pad = ignore_padding
        im = im[pad:-pad, pad:-pad]
    im = cv2.Canny(im, 127, 255, 3)
    kernel = cv2.getStructuringElement(cv2.MORPH_CROSS, (3, 3))
    im = cv2.dilate(im, kernel)

    vlines = abs(np.diff(im.mean(axis=0)))
    hlines = abs(np.diff(im.mean(axis=1)))
    return max(vlines.max(), hlines.max())


@_core.Op.using(_core.I.image)
def show_image(image):
    from matplotlib import pyplot as plt
    from PIL import Image

    if not isinstance(image, Image.Image):
        if len(image.shape) > 2:
            channels = image.shape[2]
            if channels == 4:
                image = cv2.cvtColor(image, cv2.COLOR_BGRA2RGBA)
            elif channels == 3:
                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    plt.imshow(image)
    plt.axis('off')
    plt.axis('tight')
    plt.axis('image')
    plt.show()


@_core.Op.using(_core.I.image) >> _core.I.image
def add_alpha_channel(image, alpha):
    image = cv2.cvtColor(image, cv2.COLOR_BGR2BGRA)
    alpha = np.uint8(alpha.clip(0, 255))
    alpha = cv2.resize(alpha, (image.shape[1], image.shape[0]))
    image[:, :, 3] = alpha
    return image


@_core.Op >> _core.I.image
def create_grid(*images):
    import math
    from matplotlib import pyplot as plt
    from PIL import Image
    import io

    div = math.ceil(math.sqrt(len(images)))
    for index, im in enumerate(images):
        if len(im.shape) > 2:
            channels = im.shape[2]
            if channels == 4:
                im = cv2.cvtColor(im, cv2.COLOR_BGRA2RGBA)
            elif channels == 3:
                im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)

        plt.subplot(div, div, index + 1, frameon=False, aspect='equal')
        plt.imshow(im)
        plt.axis('off')

    plt.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)
    buffer = io.BytesIO()
    plt.savefig(buffer, format='png')
    plt.close()
    buffer.seek(0)
    return Image.open(buffer)



from typing import Iterable


class Value:
    def __init__(self, value, sign=None):
        self.value = value
        self.sign = sign

    def __repr__(self):
        return repr(self.value)

    @classmethod
    def to_rel(cls, value, min_val, max_val=None):
        if isinstance(value, Iterable):
            if not isinstance(min_val, Iterable):
                min_val = [min_val] * len(value)
            if max_val is None:
                max_val = [None] * len(min_val)
            elif not isinstance(max_val, Iterable):
                max_val = [max_val] * len(value)

            return value.__class__(
                Value.to_rel(item, mn, mx)
                for item, mn, mx in zip(value, min_val, max_val)
            )
        elif isinstance(value, RelativeValue):
            return value
        else:
            if max_val is None:
                min_val, max_val = 0, min_val
            if value.sign == '+':
                max_val = max(0, max_val)
                min_val = 0
            elif value.sign == '-':
                min_val = min(0, min_val)
                max_val = 0
            return RelativeValue((value - min_val) / (max_val - min_val))

    @classmethod
    def to_abs(cls, value, min_val, max_val=None, dtype=None):
        if isinstance(value, Iterable):
            if not isinstance(min_val, Iterable):
                min_val = [min_val] * len(value)
            if max_val is None:
                max_val = [None] * len(min_val)
            elif not isinstance(max_val, Iterable):
                max_val = [max_val] * len(value)

            return value.__class__(
                Value.to_abs(item, mn, mx, dtype)
                for item, mn, mx in zip(value, min_val, max_val)
            )
        else:
            if isinstance(value, RelativeValue):
                if max_val is None:
                    min_val, max_val = 0, min_val
                if value.sign == '+':
                    max_val = max(0, max_val)
                    min_val = 0
                elif value.sign == '-':
                    min_val = min(0, min_val)
                    max_val = 0

                result = min_val + value.value * (max_val - min_val)
            else:
                result = value

            return dtype(result) if dtype else result


class RelativeValue(Value):
    @property
    def pos(self):
        return self.__class__(self.value, '+')

    @property
    def neg(self):
        return self.__class__(self.value, '-')

    def __rmod__(self, value):
        return self.__class__(value / 100., self.sign)

    def __call__(self, value, sign=...):
        return self.__class__(value, self.sign if sign == ... else sign)

    def __repr__(self):
        val = format(self.value * 100., '.10f')
        val = val.rstrip('0')
        val = val.rstrip('.')
        sign = self.sign or ''
        return f'{sign}{val}%Rel'
        #return f'{self.__class__.__name__}({self.value})'
